__author__ = "Trey Guest"
__license__ = "GNU"
__version__ = "1.0.0"
__maintainer__ = "Trey Guest"
__email__ = "twguest@students.latrobe.edu.au"
__status__ = "Production"

#    Chromosome is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    Chromosome is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with Chromosome. If not, see <http://www.gnu.org/licenses/>.

import sys
import numpy as np
import pandas as pd
import random
import datetime

from utils.progress import progress
from utils.selection_utils import Tournament

from fitness.euclid import Euclid, euclid, euclid_sn
from fitness.cosine import Cosine, cosine, cosine_sn
from fitness.dice import Dice, dice, dice_sn

from breed.npoint import Onepoint, Npoint
from breed.rand_choice import rnd_best

from selection.linrank import LinRank
from selection.rws import RoulletteWheelSelection
from selection.best import Best
from selection.tos import TournamentSelection
from selection.sus import StochasticUniversalSampling

from mutation.rnd_mutation import rnd_mutation

class Genome:

    """
    The genome object carries all genetic information for the target and seed
    matrices over any number of generations. All operations act on genome to
    progress the evolution of the population
    """


    def __init__(self, target, pop, nseeds, function, fitness = None):

        """
        :param target: target matrix object (as defined by target class)
        :param pop: optimal population size
        :param nseeds: the number of individuals generated in initio
        :param function: the transformation applied to the input matrix matrix
        :param fitness: global fitness function
        """

        self.generation = 0
        self.pop = pop # size of population
        self.nseeds = nseeds

        self.target = target
        # Here we inheret some params from the target class
        self.x = target.x
        self.y = target.y
        self.seed_dim = target.seed_dim

        if self.seed_dim == 3:
            self.z = target.z
        else:
            pass

        self.val_rng = target.val_rng
        self.val_max = target.val_max
        self.val_min = target.val_min
        self.decimals = target.decimals
        self.output = target.target

        self.function = function
        self.startTime = datetime.datetime.now()

        self.mode = ""
        self.time = 0

        self.population = []
        self.strongest = []
        self.cost = []
        self.strongest_gene = []
        self.processes = 0

        self.fitness = ""

        if fitness == None:
            self.fitness = euclid
        elif fitness == "euclid":
            self.fitness = euclid
            self.fitness_test = euclid_sn
        elif fitness == "cosine":
            self.fitness = cosine
            self.fitness_test = cosine_sn
        elif fitness == "dice":
            self.fitness = dice
            self.fitness_test = dice_sn


        # LOGGING
        self.stats_generation = []
        self.stats_strongest = []
        self.stats_winner = []
        self.stats_cost = []
        self.stats_ancestors = []
        self.stats_processes = []
        self.stats_time = []

################################################################################

    def init_population(self):

        """
        Population initialisation function

        ### Future Work should support smart population seeding - @twguest

        """

        if self.seed_dim == 3:
            x = self.x
            y = self.y
            z = self.z

            self.population = []

            i = 0

            while i in range(self.nseeds):
                if i % 5 == 0:
                    print("Initialising Population Member {} of {}".format(i, self.nseeds))
                self.population.append(np.round(np.random.uniform(self.val_min, self.val_max, [x,y,z]), self.decimals))
                i+=1

        elif self.seed_dim == 2:
            x = self.x
            y = self.y

            self.population = []

            i = 0

            while i in range(self.nseeds):
                if i % 5 == 0:
                    print("Initialising Population Member {} of {}".format(i, self.nseeds))
                self.population.append(np.round(np.random.uniform(self.val_min, self.val_max, [x,y]), self.decimals))
                i+=1

        print("************ Population Initialised")

        return self.population


################################################################################

    def display(self):

        """
        Display Utility: Flavour as neccesary
        """

        timeDiff = datetime.datetime.now() - self.startTime
        self.time = timeDiff.total_seconds()
        print("{}\t{}\t{}".format(self.generation, self.strongest, timeDiff))

################################################################################

    def linrank(self, pressure = None):

        """
        Wrapper function for Linrank
        """

        if pressure == None:
            pressure = 1

        LinRank(self)

    def RWS(self):

        """
        Wrapper function for RoulletteWheelSelection
        """

        RoulletteWheelSelection(self)

    def best(self):

        """
        Wrapper function for Best selection
        """

        Best(self)

    def TS(self, n = None):

        """
        Wrapper function for Tournament Selection
        """

        if n == None:
            n = np.ceil(int(self.pop/10))

        TournamentSelection(self, n)

    def SUS(self):

        """
        Wrapper Function for Stochastic Universal Sampling
        """

        StochasticUniversalSampling(self)

###############################################################################

    def onepoint(self, nchildren = None):

        """
        Onepoint Crossover Breeding
        : param nchildren: average number of children per individual (eff itr.)
        """

        if nchildren == None:
            nchildren = 1

        while len(self.population) < self.pop+self.pop*nchildren:
            self.population.append(Onepoint(self))

        progress(self)

        return self.population

    def npoint(self, N = None, nchildren = None):

        """ Npoint Crossover Breeding
        : param N: number of crossover points/slices
        : param nchildren: average number of children per individual (eff itr.)
        """

        if N == None:
            N = 2

        if nchildren == None:
            nchildren = 1


        while len(self.population) < self.pop+self.pop*nchildren:
            self.population.append(Npoint(self, N))

        progress(self)

        return self.population

    def rndbest(self, nI = None, nchildren = None):

        """
        A random crossover mechanism for the best N individuals
        : param nI: number of individuals
        : param nchildren: average number of children per individual (eff itr.)
        """
        if nchildren == None:
            nchildren = 1

        while len(self.population) < self.pop + self.pop*nchildren:
            self.population.append(rnd_best(self, nI))

        progress(self)
        return self.population


###############################################################################

    def random_mutation(self, nM = None, nI = None):

        """
        A random mutation function for generating nM mutations in nI
        individuals
        :param nM: number of mutations
        :param nI: number of individuals
        """

        if nM == None:
            nM = 1
        if nI == None:
            nI = 1

        itr = 0
        while itr < nI:
            rnd_mutation(self, nM)
            itr +=1

###############################################################################

if __name__ == "__main__":
    pass



__author__ = "Trey Guest"
__license__ = "GNU"
__version__ = "1.0.0"
__maintainer__ = "Trey Guest"
__email__ = "twguest@students.latrobe.edu.au"
__status__ = "Production"

#    Chromosome is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    Chromosome is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with Chromosome. If not, see <http://www.gnu.org/licenses/>.

import numpy as np

class Target:

    """
    A target class for controlling the properties of the desired output

    This target class should be defined prior to building a genome
    """

    def __init__(self,
                 target,
                 val_rng = None,
                 decimals = None,
                 exception = None,
                 shape = None):

        """
        :param target: target matrix
        :param rng: lower and upper limits [min, max] of values in trgt matrix
        :param decimals: the number of decimals for seeding/mutation
        :param shape: shape of the seed matrix [x1,y1,(z1)]

        ********** eventually i would like this to include some exceptions but
        ********** thats difficult and time consuming (ie., at this posn, use
        ********** these parameters)
        """

        self.target = target


        if shape == None:
            self.seed_dim = len(self.target.shape)

            if self.seed_dim == 2:

                self.x = self.target.shape[0]
                self.y = self.target.shape[1]

            elif self.seed_dim == 3:

                self.x = self.target.shape[0]
                self.y = self.target.shape[1]
                self.z = self.target.shape[2]

            elif self.seed_dim != 2 and self.seed_dim != 3:
                print("Your target should be a 2D or 3D array")

        elif shape != None:
            self.seed_dim = len(shape)

            if self.seed_dim == 2:
                self.x = shape[0]
                self.y = shape[1]

            if self.seed_dim == 3:
                self.x = shape[0]
                self.y = shape[1]
                self.z = shape[2]

            elif self.seed_dim != 2 and self.seed_dim != 3:
                print("Your target should be a 2D or 3D array")


        if val_rng == None:
            self.val_rng = [min(np.nditer(self.target)), max(np.nditer(self.target))]

            self.val_min = self.val_rng[0]
            self.val_max = self.val_rng[1]
        else:
            self.val_rng = [val_rng[0], val_rng[1]]

            self.val_min = val_rng[0]
            self.val_max = val_rng[1]

        if decimals == None:
            self.decimals = 3
        else:
            self.decimals = decimals


    def details(self):

        """
        print properties of target class
        """

        if len(self.target.shape) == 2:
            print("Target Size: [{},{}]".format(self.target.shape[0], self.target.shape[1]))
        elif len(self.target.shape) == 3:
            print("Target Size: [{},{},{}]".format(self.target.shape[0], self.target.shape[1], self.target.shape[2]))

        if self.seed_dim == 2:
            print("Seed Size: [{},{}]".format(self.x, self.y))
        elif self.seed_dim == 3:
            print("Seed Size: [{},{},{}]".format(self.x, self.y, self.z))

        print("Target Dimensions: {}".format(len(self.target.shape)))
        print("Seed Dimensions: {}".format(self.seed_dim))
        print("Target Value Range: {} - {}".format(self.val_min, self.val_max))
        print("Decimals: {}".format(self.decimals))





__author__ = "Trey Guest"
__license__ = "GNU"
__version__ = "1.0.0"
__maintainer__ = "Trey Guest"
__email__ = "twguest@students.latrobe.edu.au"
__status__ = "Production"

#    Chromosome is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    Chromosome is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with Chromosome. If not, see <http://www.gnu.org/licenses/>.

import pandas as pd

from matplotlib import pyplot as plt

class Darwin():

    """
    Darwin is an evolutionary companion for streamlined analysis of the
    an evolutionary algorithm
    :param genome: the genome object for analysis
    """

    def __init__(self, genome):

        self.genome = genome

        self.log = pd.DataFrame()
        self.log['Generation'] = genome.stats_generation
        self.log['Strongest'] = genome.stats_strongest
        self.log['Winner'] = genome.stats_winner
        self.log['Mean Cost'] = genome.stats_cost
        self.log['Process Count'] = genome.stats_processes
        self.log['Time'] = genome.stats_time


    def get_log(self):
        """
        Return log dataframe
        """
        return self.log


    def get_ancestors(self, generation = None):

        """
        Returns a past population
        :param generation: generation to return past population from
        """

        if generation == None:
            generation = max(self.genome.log_generation)

        if generation > max(self.genome.log_generation):
            print("Generation Should be in the range: 0-{}".format(max(self.genome.log_generation)))

        elif generation <= max(self.genome.log_generation):
            ancestors = self.genome.log_ancestors[generation]

        return ancestors



    def get_lineage(self):

        """
        Get all past populations
        """

        lineage = self.genome.stats_ancestors

        return lineage

    def plot_cost(self, save = None, save_dir = None):

        """
        A one line plot module to return the cost optimisation function
        :param save: (Bool)
        :param save_dir: desired save_dir w/ file extension
        """

        fig, ax = plt.subplots()

        plt.plot(self.log["Generation"], self.log["Strongest"], 'b')
        plt.plot(self.log["Generation"], self.log["Mean Cost"], '--b')


        ax.set_title("Cost Optimisation")

        ax.set_xlabel("Generations")
        ax.set_ylabel("Cost")

        ax.legend(["Minimum Cost", "Mean Cost"])

        if save == True:
            if save_dir == None:
                plt.savefig("ProcessControl.png")
            elif save_dir != None:
                plt.savefig(save_dir)
            elif save == False:
                pass

    def plot_processes(self, save = None, save_dir = None):

        """
        A one line plot module to return the process optimisation function
        :param save: (Bool)
        :param save_dir: desired save_dir w/ file extension
        """

        fig, ax = plt.subplots(1, 3, figsize = [29,7])

        fig.suptitle("Process Report", fontsize = 18)

        plt.subplot(131)
        plt.plot(self.log["Generation"], self.log["Process Count"])
        plt.title("Generation")
        plt.xlabel("Generations")
        plt.ylabel("Processes")

        plt.subplot(132)
        plt.plot(self.log["Process Count"], self.log["Strongest"], 'b')
        plt.plot(self.log["Process Count"], self.log["Mean Cost"], '--b')
        plt.title("Cost Optimisation")
        plt.xlabel("Processes")
        plt.ylabel("Cost")
        plt.legend(["Minimum Cost", "Maximum Cost"])

        plt.subplot(133)
        plt.plot(self.log["Time"], self.log["Process Count"])
        plt.title("Time of Operation")
        plt.xlabel("Time (s)")
        plt.ylabel("Processes")

        if save == True:
            if save_dir == None:
                plt.savefig("ProcessControl.png")
            elif save_dir != None:
                plt.savefig(save_dir)
            elif save == False:
                pass

if __name__ == "__main__":
    pass



#!

from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='Chromosome',
      version='0.1.1',
      description='Evolutionary Algorithm Framework in Python',
      author='Trey Guest',
      author_email='twguest@students.latrobe.edu.au',
      url='https://www.github.com/twguest/Chromosome',
      packages=find_packages(exclude=['examples']),
      platforms=['any'],
      keywords=['evolutionary algorithms','genetic algorithms','genetic programming','cma-es','ga','gp','es','pso', 'artificial intelligence', 'ai'],
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='LGPL',
      classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        ],
     use_2to3=True
)
